\ MSP430F149 small-legs control for hexapod beatle
\ The (GO) command puts the legs in motion
\ So 10 20 15 leg1 (go) moves all joints of leg1
\ h = horizontal,  s = shoulder,  e = elbow 
\ The h(orizontal) range from 70 degrees backward to 110 degrees forward
\ So the most forward legs can maybe be used to drag something away
\ and using touch to explore the world??
\ The s(houlder) moves from 90 degrees upward to 90 degrees downward
\ The e(lbow) moves 90 degrees square to both sides of the s(houlder)

\ ---\      ---\       ---|      ---|     ---|     ---|---|
\     \         --        |         |        |            |
\     |          |     ---         /        /             |
\     |                          /         | 
\   Norm       Up      Rest     Down     Push        Help
\ 80-60-D0  80-A0-D0  80-FF-FF  80-FF-D0  80-FF-C0  80-80-FF
\
\   p6 ------- p1    The legs are attached to the beatle this way!
\      |     |
\      |     |
\   p5 |     | p2
\      |     |
\      |     |
\   p4 ------- p3

\ Note that this is all experimental
\ but a real 6 leg creature has arrived!
\ Just checking real positions now.

\ Next steps:
\ Turn to left and right..
\ Walk like a crab (sidewards)..
\ Big and small steps
\ Running, etc.
inside
: SPEED@    wait @ ;
forth
: (HEAD)    00 (joint)  01 (joint) ;
: HEAD      (head) (go) ;

\ Invert servo movement for the other side, so both sides
\ can use the same numbers to move joints.
code INV    ( b1 -- b2 )
    #-1 day mov
    tos day sub
    #-1 day .b and>
    #0 day cmp <>? if,
        #-1 day .b cmp <>? if,
            #1 day add
        then,
    then,
    day tos mov
    next
end-code

: LEG1      02 (joint)  03 (joint)  04 (joint) ;  ( h s e -- )
: LEG2      05 (joint)  06 (joint)  07 (joint) ;
: LEG3      08 (joint)  09 (joint)  0a (joint) ;
: LEG4      inv 0B (joint)  inv 0c (joint)  inv 0d (joinT) ;
: LEG5      inv 0E (joint)  inv 0f (joint)  inv 10 (joint) ;
: LEG6      inv 11 (joint)  inv 12 (joint)  inv 13 (joint) ;

\ Fetch data for one leg from EEPROM
: @LEG      ec@  nec@  nec@ ;  ( -- h s e )

\ Data structures for leg movement in EEPROM
ecreate NORM      80 ec, 60 ec, D0 ec,
ecreate UP        80 ec, 90 ec, FF ec,
ecreate BACKW     90 ec, 60 ec, D0 ec,
ecreate FORW      70 ec, 60 ec, D0 ec,

ecreate DOWN      80 ec, 00 ec, D0 ec,
ecreate REST      80 ec, 00 ec, FF ec,
ecreate PUSH      80 ec, 40 ec, FF ec,
ecreate HELP      80 ec, 80 ec, FF ec,

ecreate RESTa     80 ec, 80 ec, 80 ec,
ecreate UPa       80 ec, B0 ec, FF ec,

: 3DUP      2 pick  2 pick  2 pick ;
: LLEGS     3dup leg1  3dup leg3  leg5 ;
: RLEGS     3dup leg2  3dup leg4  leg6 ;

value POS
: LEFT-NORM   ( -- )  norm @leg llegs ;
: RIGHT-NORM  ( -- )  norm @leg rlegs ;
: LEFT-UP     ( -- )  up @leg llegs ;
: RIGHT-UP    ( -- )  up @leg rlegs ;
: START       ( -- )  0 to pos  left-norm  right-norm  (go) ; \ norm position

: SHAKE       ( -- )
    speed@  0 speed  
    upa @leg leg1 (go)  norm @leg leg1 (go) 
    upa @leg leg6 (go)  norm @leg leg6 (go) 
    upa @leg leg2 (go)  norm @leg leg2 (go) 
    upa @leg leg5 (go)  norm @leg leg5 (go) 
    upa @leg leg3 (go)  norm @leg leg3 (go) 
    upa @leg leg4 (go)  norm @leg leg4 (go)
    speed ;

\ Leg  1, 3 and 5
: FSTEP1      ( -- )
    1 to pos  left-up (go)  right-norm  forw @leg llegs  (go) ;
    
\ Leg  2, 4 and 6
: FSTEP2      ( -- )
    -1 to pos  right-up (go)  left-norm  forw @leg rlegs  (go) ;

\ Leg  1, 3 and 5
: BSTEP1      ( -- )
    1 to pos  left-up (go)  right-norm  backw @leg llegs  (go) ;
    
\ Leg  2, 4 and 6
: BSTEP2      ( -- )
    -1 to pos  right-up (go)  left-norm  backw @leg rlegs  (go) ;


: >NORM       ( -- )
    pos ?dup if
        0 to pos
        0< if  left-up (go)  left-norm (go) exit  then
        right-up (go)  right-norm (go)
    then ;

: FSTEPS      ( u -- )    0 ?do  fstep1 fstep2  loop  pos . >norm ;
: BSTEPS      ( u -- )    0 ?do  bstep1 bstep2  loop  pos . >norm ;


\ Other experimental movements...

\ Fold your legs under your body (like a cat)
: SLEEP       ( -- )
    left-up (go)  down @leg llegs (go)
    right-up (go)  down @leg rlegs  (go) 
    rest @leg llegs  rest @leg rlegs  (go) ;

\ Stand up carefully from sleep position
: WAKEUP      ( -- )
    down @leg llegs  down @leg rlegs  (go)  \ First push
\   push @leg llegs  push @leg rlegs  (go)  \ Second push
    rest @leg llegs  (go)  help @leg llegs  (go) \ Free legs
    norm @leg llegs  (go)                   \ Left normal and
    rest @leg rlegs  (go)  help @leg rlegs  \ Free legs
    (go)  start ;                           \ right normal


\ Turn on your place to the left
: LSTEP       ( -- )
    right-up (go)  left-norm  forw @leg leg4  
    forw @leg leg6  backw @leg leg2  (go)
    left-up (go)  right-norm  backw @leg leg1  
    backw @leg leg3  forw @leg leg5  (go) ;


\ Crab like walk... to the right
ecreate CRAB1     80 ec, 60 ec, D0 ec, 
ecreate CRAB2     80 ec, 60 ec, FF ec, 
ecreate UPCRAB    80 ec, 80 ec, FF ec, 

: CRAB-REST   ( -- )    
    upcrab @leg rlegs (go)  crab1 @leg rlegs crab1 @leg llegs (go) ;

: CRAB        ( -- )
    right-up (go)
    crab2 @leg leg1  crab2 @leg leg3  crab1 @leg leg5 (go)
\    crab1 @leg rlegs (go)  
    left-up (go)
    crab1 @leg leg2  down @leg leg6  down @leg leg4 (go) ;

: LTURN       ( u -- )  0 ?do  lstep  loop  shake ;
: RCRAB       ( u -- )  0 ?do  crab  loop  crab-rest ;

: REST1 ( -- )  resta @leg 3dup llegs rlegs (go) ;
: UP1   ( -- )  norm @leg 3dup llegs rlegs (go) shake ;
: UP2   ( -- )  upa @leg 3dup llegs rlegs (go) up1 ;

inside
: READY         ( -- )
    servo-on  ta-on 200 ms  tb-on 200 ms  zero-servos  
    ( setup-pili ) erase-all   40 speed  up2 ;
forth

ecreate LSTEPF  50 ec,  60 ec,  D0 ec,
ecreate LSTEPB  B0 ec,  60 ec,  D0 ec,
ecreate LUP     B0 ec,  A8 ec,  FF ec,

: FW            ( -- )
    lup @leg leg1 (go)  lstepf @leg leg1 (go)
    lup @leg leg6 (go)  lstepf @leg leg6 (go)
    lup @leg leg2 (go)  lstepf @leg leg2 (go)
    lup @leg leg5 (go)  lstepf @leg leg5 (go)
    lup @leg leg3 (go)  lstepf @leg leg3 (go)
    lup @leg leg4 (go)  lstepf @leg leg4 (go)
    lstepb @leg  3dup llegs rlegs (go) ;

: CENTI         ( -- )   
    lstepb @leg  3dup llegs rlegs (go)
    0 ?do  fw  loop  start ;

shield legs\  freeze

\ End
